//----------------------------------------------------------------------------
//
// LFSR functions with x86 optimizations, by Scott Duplichan
// tested with Microsoft and mingw C compilers
//
//----------------------------------------------------------------------------
#include "project.h"
//----------------------------------------------------------------------------

#if defined (_MSC_VER) && (_MSC_VER < 1700)
void lfsrYmm (TEST_DATA *testData, int loops, uint64_t cpuSpeed) {}
#else

//----------------------------------------------------------------------------
// bit shift left a 256-bit value using ymm registers
//          __m256i *data - data to shift
//          int count     - number of bits to shift
// return:  __m256i       - carry out bit(s)

static __m256i bitShiftLeft256ymm (__m256i *data, int count)
   {
   __m256i innerCarry, carryOut, rotate;

   innerCarry = _mm256_srli_epi64 (*data, 64 - count);                        // carry outs in bit 0 of each qword
   rotate     = _mm256_permute4x64_epi64 (innerCarry, 0x93);                  // rotate ymm left 64 bits
   innerCarry = _mm256_blend_epi32 (_mm256_setzero_si256 (), rotate, 0xFC);   // clear lower qword
   *data    = _mm256_slli_epi64 (*data, count);                               // shift all qwords left
   *data    = _mm256_or_si256 (*data, innerCarry);                            // propagate carrys from low qwords
   carryOut   = _mm256_xor_si256 (innerCarry, rotate);                        // clear all except lower qword
   return carryOut;
   }

//----------------------------------------------------------------------------
// bit shift left a 512-bit value using ymm registers
//          __m256i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m256i       - carry out bit(s)

static __m256i bitShiftLeft512ymm (__m256i *data, int count)
   {
   __m256i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft256ymm (&data [0], count);
   carryOut1 = bitShiftLeft256ymm (&data [1], count);
   data [1] = _mm256_or_si256 (data [1], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
// bit shift left a 1024-bit value using ymm registers
//          __m256i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m256i       - carry out bit(s)

static __m256i bitShiftLeft1024ymm (__m256i *data, int count)
   {
   __m256i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft512ymm (&data [0], count);
   carryOut1 = bitShiftLeft512ymm (&data [2], count);
   data [2] = _mm256_or_si256 (data [2], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
// bit shift left a 2048-bit value using ymm registers
//          __m256i *data - data to shift, ls part stored first 
//          int count     - number of bits to shift
// return:  __m256i       - carry out bit(s)

static __m256i bitShiftLeft2048ymm (__m256i *data, int count)
   {
   __m256i carryOut0, carryOut1;

   carryOut0 = bitShiftLeft1024ymm (&data [0], count);
   carryOut1 = bitShiftLeft1024ymm (&data [4], count);
   data [4] = _mm256_or_si256 (data [4], carryOut0);
   return carryOut1;
   }

//----------------------------------------------------------------------------
//
// lfsr256ymm - left shift galois type lfsr for 256-bit data, ymm implementation
//

static __m256i lfsr256ymm (__m256i data, __m256i mask)
   {
   __m256i maskOrZero;

   // expand data bit 255 into mask 0-255
   maskOrZero = _mm256_permute4x64_epi64 (data, 0xEE);      // copy high half to low half
   maskOrZero = _mm256_shuffle_epi32 (maskOrZero, 0xFF);    // copy high dword to all dwords
   maskOrZero = _mm256_srai_epi32 (maskOrZero, 31);         // all zeros or all ones

   // zero the mask if data bit 255 is zero
   maskOrZero = _mm256_and_si256 (maskOrZero, mask);

   // shift the whole data left by one bit
   bitShiftLeft256ymm (&data, 1);

   // xor in mask if bit 255 was set
   return _mm256_xor_si256 (data, maskOrZero);
   }

//----------------------------------------------------------------------------
//
// lfsr512ymm - left shift galois type lfsr for 512-bit data, ymm implementation
//

static void lfsr512ymm (__m256i *data, __m256i *mask)
   {
   __m256i maskOrZero [2], maskOrZ;

   // expand data bit 511 into mask 0-511
   maskOrZ = _mm256_permute4x64_epi64 (data [1], 0xEE);     // copy high half to low half
   maskOrZ = _mm256_shuffle_epi32 (maskOrZ, 0xFF);          // copy high dword to all dwords
   maskOrZ = _mm256_srai_epi32 (maskOrZ, 31);               // all zeros or all ones

   // zero the mask if data bit 511 is zero
   maskOrZero [0] = _mm256_and_si256 (maskOrZ, mask[0]);
   maskOrZero [1] = _mm256_and_si256 (maskOrZ, mask[1]);

   // shift the whole data left by one bit
   bitShiftLeft512ymm (data, 1);

   // xor in mask if bit 511 was set
   data [0] = _mm256_xor_si256 (data [0], maskOrZero [0]);
   data [1] = _mm256_xor_si256 (data [1], maskOrZero [1]);
   }

//----------------------------------------------------------------------------
//
// lfsr1024ymm - left shift galois type lfsr for 1024-bit data, ymm implementation
//

static void lfsr1024ymm (__m256i *data, __m256i *mask)
   {
   __m256i maskOrZero [2], maskOrZ;

   // expand data bit 1023 into mask 0-1023
   maskOrZ = _mm256_permute4x64_epi64 (data [3], 0xEE);     // copy high half to low half
   maskOrZ = _mm256_shuffle_epi32 (maskOrZ, 0xFF);          // copy high dword to all dwords
   maskOrZ = _mm256_srai_epi32 (maskOrZ, 31);               // all zeros or all ones

   // zero the mask if data bit 1023 is zero
   maskOrZero [0] = _mm256_and_si256 (maskOrZ, mask[0]);
   maskOrZero [1] = _mm256_and_si256 (maskOrZ, mask[1]);
   maskOrZero [2] = _mm256_and_si256 (maskOrZ, mask[2]);
   maskOrZero [3] = _mm256_and_si256 (maskOrZ, mask[3]);

   // shift the whole data left by one bit
   bitShiftLeft1024ymm (data, 1);

   // xor in mask if bit 1023 was set
   data [0] = _mm256_xor_si256 (data [0], maskOrZero [0]);
   data [1] = _mm256_xor_si256 (data [1], maskOrZero [1]);
   data [2] = _mm256_xor_si256 (data [2], maskOrZero [2]);
   data [3] = _mm256_xor_si256 (data [3], maskOrZero [3]);
   }

//----------------------------------------------------------------------------
//
// lfsr2048ymm - left shift galois type lfsr for 2048-bit data, ymm implementation
//
static void lfsr2048ymm (__m256i *data, __m256i *mask)
   {
   __m256i maskOrZero [2], maskOrZ;

   // expand data bit 2047 into mask 0-2047
   maskOrZ = _mm256_permute4x64_epi64 (data [7], 0xEE);     // copy high half to low half
   maskOrZ = _mm256_shuffle_epi32 (maskOrZ, 0xFF);          // copy high dword to all dwords
   maskOrZ = _mm256_srai_epi32 (maskOrZ, 31);               // all zeros or all ones

   // zero the mask if data bit 2047 is zero
   maskOrZero [0] = _mm256_and_si256 (maskOrZ, mask[0]);
   maskOrZero [1] = _mm256_and_si256 (maskOrZ, mask[1]);
   maskOrZero [2] = _mm256_and_si256 (maskOrZ, mask[2]);
   maskOrZero [3] = _mm256_and_si256 (maskOrZ, mask[3]);
   maskOrZero [4] = _mm256_and_si256 (maskOrZ, mask[4]);
   maskOrZero [5] = _mm256_and_si256 (maskOrZ, mask[5]);
   maskOrZero [6] = _mm256_and_si256 (maskOrZ, mask[6]);
   maskOrZero [7] = _mm256_and_si256 (maskOrZ, mask[7]);

   // shift the whole data left by one bit
   bitShiftLeft2048ymm (data, 1);

   // xor in mask if bit 2047 was set
   data [0] = _mm256_xor_si256 (data [0], maskOrZero [0]);
   data [1] = _mm256_xor_si256 (data [1], maskOrZero [1]);
   data [2] = _mm256_xor_si256 (data [2], maskOrZero [2]);
   data [3] = _mm256_xor_si256 (data [3], maskOrZero [3]);
   data [4] = _mm256_xor_si256 (data [4], maskOrZero [4]);
   data [5] = _mm256_xor_si256 (data [5], maskOrZero [5]);
   data [6] = _mm256_xor_si256 (data [6], maskOrZero [6]);
   data [7] = _mm256_xor_si256 (data [7], maskOrZero [7]);
   }

//===========================================================================
//
// measure execution time and verify proper operation for each lfsr function
//

void lfsrYmm (TEST_DATA *testData, int loops, uint64_t cpuSpeed)
   {
   int index;
   uint64_t startCount;
   static __m256i  pattern256, pattern512 [2], pattern1024 [4], pattern2048 [8];
   static __m256i  mask256   , mask512    [2], mask1024    [4], mask2048    [8];

   memcpy (&pattern256, testData->initial , sizeof pattern256 );
   memcpy (pattern512 , testData->initial , sizeof pattern512 );
   memcpy (pattern1024, testData->initial , sizeof pattern1024);
   memcpy (pattern2048, testData->initial , sizeof pattern2048);
   memcpy (&mask256   , testData->mask256 , sizeof mask256    );
   memcpy (mask512    , testData->mask512 , sizeof mask512    );
   memcpy (mask1024   , testData->mask1024, sizeof mask1024   );
   memcpy (mask2048   , testData->mask2048, sizeof mask2048   );

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) pattern256 = lfsr256ymm (pattern256, mask256);

   printf ("lfsr256ymm  %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (&pattern256, testData->expected256, sizeof pattern256) != 0) printf ("-----lfsr256ymm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr512ymm (pattern512, mask512);

   printf ("lfsr512ymm  %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern512, testData->expected512, sizeof pattern512) != 0) printf ("-----lfsr512ymm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr1024ymm (pattern1024, mask1024);

   printf ("lfsr1024ymm %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern1024, testData->expected1024, sizeof pattern1024) != 0) printf ("-----lfsr1024ymm test failed-----\n");

   //----------------------------------------------------------------------------

   startCount = queryPerformanceCounter ();
   for (index = 0; index < loops; index++) lfsr2048ymm (pattern2048, mask2048);

   printf ("lfsr2048ymm %s\n", cyclesText (queryPerformanceCounter () - startCount, cpuSpeed, loops));
   if (memcmp (pattern2048, testData->expected2048, sizeof pattern2048) != 0) printf ("-----lfsr2048ymm test failed-----\n");

   //----------------------------------------------------------------------------
   }
#endif